/*===================================================================================================================================================================================================================================
                                                                                                [Подгружаем API amxmodx]
===================================================================================================================================================================================================================================*/

#include <amxmodx>

/*===================================================================================================================================================================================================================================
                                                                                                [Макросы]
===================================================================================================================================================================================================================================*/

#define TutorSoundNone        -1

#define MAX_MSG_LEN            369
#define MAX_MSG_LEN2        185

#define PRECACHE_SOUND        precache_sound
#define PRECACHE_GENERIC    precache_generic
#define GET_USER_MSG_ID        get_user_msgid
#define MESSAGE_BEGIN        message_begin
#define WRITE_BYTE            write_byte
#define WRITE_SHORT            write_short
#define WRITE_STRING        write_string
#define MESSAGE_END()        message_end()

/*===================================================================================================================================================================================================================================
                                                                                                [Переменные]
===================================================================================================================================================================================================================================*/

// Целочисленные
new g_iMsgTutorClose;    // http://wiki.alliedmods.net/Half-Life_1_Game_Events#TutorClose    No arguments
new g_iMsgTutorText;    // http://wiki.alliedmods.net/Half-Life_1_Game_Events#TutorText        string byte string string short short short
new g_iMaxPlayers;
static const g_iSounds[][] =
{
    "events/enemy_died.wav",
    "events/friend_died.wav",
    "events/task_complete.wav",
    "events/tutor_msg.wav"
};

// Булевая
new bool:g_bReliable;    // Надёжность

// Динамический массив
new Array:g_aMultiLang;    // Для записи мест которые нужно менять (ML)

/*===================================================================================================================================================================================================================================
                                                                                                [Natives, Precache, Init, End]
===================================================================================================================================================================================================================================*/

// Natives
public plugin_natives()
{
    register_native("TutorReliable", "NativeTutorReliable", 1);    // Изменение надёжности
    register_native("TutorShow", "NativeTutorShow", 0);            // Показ
    register_native("TutorClose", "NativeTutorClose", 1);        // Закрытие
}

// Precache
public plugin_precache()
{
    // Graffic Effects
	PRECACHE_GENERIC("gfx/career/icon_i.tga");
    PRECACHE_GENERIC("gfx/career/icon_skulls.tga");
    PRECACHE_GENERIC("gfx/career/round_corner_ne.tga");
    PRECACHE_GENERIC("gfx/career/round_corner_nw.tga");
    PRECACHE_GENERIC("gfx/career/round_corner_se.tga");
    PRECACHE_GENERIC("gfx/career/round_corner_sw.tga");
    
    // Resources
    PRECACHE_GENERIC("resource/TutorScheme.res");
    PRECACHE_GENERIC("resource/UI/TutorTextWindow.res");
    
    // Resources
    PRECACHE_GENERIC("resource/TutorScheme.res");
    PRECACHE_GENERIC("resource/UI/TutorTextWindow.res");
    
    // Sounds
    PRECACHE_SOUND("events/enemy_died.wav");
    PRECACHE_SOUND("events/friend_died.wav");
    PRECACHE_SOUND("events/task_complete.wav");
    PRECACHE_SOUND("events/tutor_msg.wav");
}

// Init
public plugin_init()
{
    // Регистрируем
    register_plugin("Tutor API", "Alpha", "[WPMG]PRoSToTeM@");
    
    // Получаем идентификаторы сообщений
    g_iMsgTutorClose = GET_USER_MSG_ID("TutorClose");
    g_iMsgTutorText = GET_USER_MSG_ID("TutorText");
    
    // Получаем максимальное количество игроков
    g_iMaxPlayers = get_maxplayers();
    
    // Создаём динамический массив
    g_aMultiLang = ArrayCreate();
}

// End
public plugin_end() ArrayDestroy(g_aMultiLang); // Удаляем массив

/*===================================================================================================================================================================================================================================
                                                                                                [AMXX API функции]
===================================================================================================================================================================================================================================*/

// Событие выхода игрока с сервера
public client_disconnect(iEntity) remove_task(iEntity); // Убираем задачу

/*===================================================================================================================================================================================================================================
                                                                                                [Native handlers]
===================================================================================================================================================================================================================================*/

// TutorReliable(bool:bReliable = false)
// Изменяет надёжность сообщения тутора
public NativeTutorReliable(bool:bReliable) g_bReliable = bReliable;

// TutorShow(iEntity = 0, iType = TutorDefault, Float:fTime = 0.0, iSound = TutorSoundNone, const iMessage[], any:...)
// Показываем тутор
public NativeTutorShow(iPlugin, iParams)
{
    // Получаем значения аргументов
    new iEntity = get_param(1);            // 1-ый
    new iType = get_param(2);            // 2-ой
    new Float:fTime = get_param_f(3);    // 3-ий
    new iSound = get_param(4);            // 4-ый
    
    // Сообщение
    new iBuffer[MAX_MSG_LEN];
    new iBuffer2[MAX_MSG_LEN2];
    
    // Отсылаем одному игроку
    if (iEntity)
    {
        // Заменяем все LANG_PLAYER, на ID получателя
        for (new iArg = 6; iArg < iParams; iArg++)
        {
            // Если аргумент не равен LANG_PLAYER переходим к следующему аргументу
            if (get_param_byref(iArg) != LANG_PLAYER)
                continue;
            
            // Получаем следующий аргумент
            get_string(iArg + 1, iBuffer, charsmax(iBuffer));
            
            // Если это не ML ключ переходим к следующему аргументу
            if (GetLangTransKey(iBuffer) == TransKey_Bad)
                continue;
            
            // Меняем численное значение аргумента
            set_param_byref(iArg++, iEntity);
        }
        
        // Если аргументов больше 5 форматируем сообщение
        if (iParams > 5) vdformat(iBuffer, charsmax(iBuffer), 5, 6);
        // Иначе получаем 5 аргумент
        else get_string(5, iBuffer, charsmax(iBuffer));
        
        // Преобразуем
        utf8Tocp1251(iBuffer, iBuffer2, charsmax(iBuffer2));
        
        // Отсылаем
        MESSAGE_BEGIN(g_bReliable ? MSG_ONE : MSG_ONE_UNRELIABLE, g_iMsgTutorText, _, iEntity);
        WRITE_STRING(iBuffer2);
        WRITE_BYTE(0);
        WRITE_SHORT(0);
        WRITE_SHORT(0);
        WRITE_SHORT(iType);
        MESSAGE_END();
        
        // Убираем task
        remove_task(iEntity);
        
        // Если есть время, то задаём задачу
        if (fTime)
        {
            // Записываем в аргументы задачи надёжность
            new iTaskArgs[1];
            
            iTaskArgs[0] = _:g_bReliable;
            
            set_task(fTime, "TaskTutorClose", iEntity, iTaskArgs, 1);
        }
        
        // Проигрываем звук
        if (iSound != TutorSoundNone) PlaySound(iEntity, _, g_iSounds[iSound], _, g_bReliable);
    }
    // Отсылаем всем
    else
    {
        // Количество аргументов, текущий аргумент
        new iChangeArgsCount, iArg;
        
        // Записываем все аргументы содержащие LANG_PLAYER
        for (iArg = 6; iArg < iParams; iArg++)
        {
            // Если аргумент не равен LANG_PLAYER переходим к следующему аргументу
            if (get_param_byref(iArg) != LANG_PLAYER)
                continue;
            
            // Получаем следующий аргумент
            get_string(iArg + 1, iBuffer, charsmax(iBuffer));
            
            // Если это не ML ключ переходим к следующему аргументу
            if (GetLangTransKey(iBuffer) == TransKey_Bad)
                continue;
            
            // Записываем в массив
            ArrayPushCell(g_aMultiLang, iArg++);
            
            iChangeArgsCount++;
        }
        
        if (iChangeArgsCount)
        {
            // Массив из игроков
            new iPlayers[32];
            // Всего игроков
            new iPlayersCount;
            // Последний игрок
            new iPlayer;
            
            // Получаем игроков
            get_players(iPlayers, iPlayersCount);
            
            // Делаем цикл из игроков
            for (new i = 0; i < iPlayersCount; i++)
            {
                // Сохраняем последнего игрока
                iPlayer = iPlayers[i];
                
                // Меняем аргументы LANG_PLAYER на ID игрока
                for (iArg = 0; iArg < iChangeArgsCount; iArg++)
                    set_param_byref(ArrayGetCell(g_aMultiLang, iArg), iPlayer);
                
                // Форматируем сообщение
                vdformat(iBuffer, charsmax(iBuffer), 5, 6);
                
                // Преобразуем
                utf8Tocp1251(iBuffer, iBuffer2, charsmax(iBuffer2));
                
                // Отсылаем
                MESSAGE_BEGIN(g_bReliable ? MSG_ONE : MSG_ONE_UNRELIABLE, g_iMsgTutorText, _, iPlayer);
                WRITE_STRING(iBuffer2);
                WRITE_BYTE(0);
                WRITE_SHORT(0);
                WRITE_SHORT(0);
                WRITE_SHORT(iType);
                MESSAGE_END();
                
                // Убираем task
                remove_task(iPlayer);
                
                // Если есть время, то задаём задачу
                if (fTime)
                {
                    // Записываем в аргументы задачи надёжность
                    new iTaskArgs[1];
                    
                    iTaskArgs[0] = _:g_bReliable;
                    
                    set_task(fTime, "TaskTutorClose", iPlayer, iTaskArgs, 1);
                }
                
                // Проигрываем звук
                if (iSound != TutorSoundNone) PlaySound(iPlayer, _, g_iSounds[iSound], _, g_bReliable);
            }
            
            // Очищаем массив
            ArrayClear(g_aMultiLang);
        }
        else
        {
            // Если аргументов больше 5 форматируем сообщение
            if (iParams > 5) vdformat(iBuffer, charsmax(iBuffer), 5, 6);
            // Иначе получаем 5 аргумент
            else get_string(5, iBuffer, charsmax(iBuffer));
            
            // Преобразуем
            utf8Tocp1251(iBuffer, iBuffer2, charsmax(iBuffer2));
            
            // Отсылаем
            MESSAGE_BEGIN(g_bReliable ? MSG_ALL : MSG_BROADCAST, g_iMsgTutorText);
            WRITE_STRING(iBuffer2);
            WRITE_BYTE(0);
            WRITE_SHORT(0);
            WRITE_SHORT(0);
            WRITE_SHORT(iType);
            MESSAGE_END();
            
            // Массив из игроков
            new iPlayers[32];
            // Всего игроков
            new iPlayersCount;
            // Последний игрок
            new iPlayer;
            
            // Получаем игроков
            get_players(iPlayers, iPlayersCount);
            
            // Делаем цикл из игроков
            for (new i = 0; i < iPlayersCount; i++)
            {
                // Сохраняем последнего игрока
                iPlayer = iPlayers[i];
                
                // Убираем task
                remove_task(iPlayer);
                
                // Если есть время, то задаём задачу
                if (fTime)
                {
                    // Записываем в аргументы задачи надёжность
                    new iTaskArgs[1];
                    
                    iTaskArgs[0] = _:g_bReliable;
                    
                    set_task(fTime, "TaskTutorClose", iPlayer, iTaskArgs, 1);
                }
            }
            
            // Проигрываем звук
            if (iSound != TutorSoundNone) PlaySound(_, _, g_iSounds[iSound], _, g_bReliable);
        }
    }
}

// TutorClose(iEntity = 0)
// Закрытие тутора
public NativeTutorClose(iEntity)
{
    if (iEntity)
    {
        MESSAGE_BEGIN(g_bReliable ? MSG_ONE : MSG_ONE_UNRELIABLE, g_iMsgTutorClose, _, iEntity);
        MESSAGE_END();
        
        remove_task(iEntity);
    }
    else
    {
        MESSAGE_BEGIN(g_bReliable ? MSG_ALL : MSG_BROADCAST, g_iMsgTutorClose);
        MESSAGE_END();
        
        for (new i = 1; i <= g_iMaxPlayers; i++)
            remove_task(i);
    }
}

/*===================================================================================================================================================================================================================================
                                                                                                [Tasks handlers]
===================================================================================================================================================================================================================================*/

public TaskTutorClose(const iTaskArgs[], iEntity)
{
    MESSAGE_BEGIN(iTaskArgs[0] ? MSG_ONE : MSG_ONE_UNRELIABLE, g_iMsgTutorClose, _, iEntity);
    MESSAGE_END();
}

/*===================================================================================================================================================================================================================================
                                                                                                [Stocks]
===================================================================================================================================================================================================================================*/

// UTF8 to cp1251 stock by PomanoB, other code styling xD
stock utf8Tocp1251(const string[], output[], maxlen)
{
    new i, len, j, char1, char2
    len = strlen(string)
    while(string[i] && j <= maxlen)
    {
        if (i + 1 < len)
        {
            char1 = string[i]&0xFF
            char2 = string[i+1]&0xFF
            if (char1 == 0xD0 && char2 == 0x81)
            {
                output[j] = 168
                i++
            }
            else
            if (char1 == 0xD1 && char2 == 0x91)
            {
                output[j] = 184
                i++
            }
            else
            if (char1 == 0xD0 && char2 >= 0x90 && char2 <= 0xBF)
            {
                output[j] = char2 + 48
                i++
            }
            else
            if (char1 == 0xD1 && char2 >= 0x80 && char2 <= 0x8F)
            {
                output[j] = char2 + 112
                i++
            }
            else
                output[j] = string[i]
        }
        else
            output[j] = string[i]
        i++
        j++
    }
    
    output[maxlen] = 0;
}

stock PlaySound(const pReceiver = 0, const pSender = 0, const szSound[], const iPitch = PITCH_NORM, const bool:bReliable = false)
{
    if (bReliable)
    {
        if (pReceiver)
        {
            message_begin(MSG_ONE, get_user_msgid("SendAudio"), _, pReceiver)
        }
        else
        {
            message_begin(MSG_ALL, get_user_msgid("SendAudio"))
        }
    }
    else
    {
        if (pReceiver)
        {
            message_begin(MSG_ONE_UNRELIABLE, get_user_msgid("SendAudio"), _, pReceiver)
        }
        else
        {
            message_begin(MSG_BROADCAST, get_user_msgid("SendAudio"))
        }
    }
    
    write_byte(pSender)
    write_string(szSound)
    write_short(iPitch)
    message_end()
}